/* Includes mpi.h, along with threadings stuff and defines internal globals. */

#ifndef __MPII_H__
#define __MPII_H__

#include "threading.h"
#define MPI_INTERNAL
#include "mpi.h"

#undef MPII_Stack_size
#define MPII_Stack_size MPII_Stack_size_val
#undef MPII_Num_proc
#define MPII_Num_proc MPII_Num_proc_val
#undef MPI_REQUEST_NULL
#define MPI_REQUEST_NULL MPII_Request_null_val

/* Structures */

struct MPII_Member_STRUCT;
typedef struct
{
   enum {MSG_AVAIL, TOOK_MSG} type;
   MPI_Request *req;
} MPII_Msg;

typedef struct
{
   MPII_Msg *q;
   int *freelist;
   int *next;
   int max, nfree, head, tail, dirty, dirty_prev;
} MPII_Msg_queue;

typedef struct MPII_Member_STRUCT
{
   /* Thread thread; */
   Mutex mutex;
   Cond cond;
   MPII_Msg_queue queue;
} MPII_Member;

/* Errors */
/* MPI_SUCCESS (0) is an error in addition to an error class (see mpi.h) */
#define MPII_NOT_INITIALIZED 1
#define MPII_BAD_ERRORCODE   2
#define MPII_OUT_OF_MEMORY   3
#define MPII_INIT_ARG        4
#define MPII_ACTIVE_REQ      5
#define MPII_NULL_REQ        6
#define MPII_RANK_RANGE      7
#define MPII_TYPE_MISMATCH   8
#define MPII_OVERFLOW        9
#define MPII_INACTIVE_REQ   10
#define MPII_MAX_ERROR      10

/* Tags */
#define MPII_BCAST_TAG     -10
#define MPII_BARRIER_TAG1  -11
#define MPII_BARRIER_TAG2  -12

#define mymalloc(comm,num,type) \
   ((type *) MPII_Malloc (comm, (num) * sizeof (type)))

/* Used for inter-thread communication */
#define post_send(target_in,request,msgvar,targetvar) \
      targetvar = target_in; \
      msgvar.type = MSG_AVAIL; \
      msgvar.req = request; \
      lock ((targetvar)->mutex); \
         MPII_enqueue (&((targetvar)->queue), &(msgvar)); \
         notify ((targetvar)->cond); \
      unlock ((targetvar)->mutex)
#define post_recv() 
#define notify_sender(src_in,msgvar,srcvar) \
      srcvar = src_in; \
      msgvar.type = TOOK_MSG; \
      /* msgvar.req = request; */ \
      lock ((srcvar)->mutex); \
         MPII_enqueue (&((srcvar)->queue), &(msgvar)); \
         notify ((srcvar)->cond); \
      unlock ((srcvar)->mutex)

/* Debug flags */
#define DEBUG_QUEUE_SEARCH 0
   /* 1 ==> show matches; 2 ==> show traversal through queue;
    * 3 ==> show dirty info
    */
#define DEBUG_ENQUEUE 0
   /* 1 ==> show enqueues; 2 ==> say where it went */

#include "iprotos.h"

#endif
